from __future__ import annotations

import abc
import logging
from contextlib import nullcontext
from typing import TYPE_CHECKING, Any, ClassVar, Generic

from typing_extensions import TypeVar, override

from waku.eventsourcing._retry import execute_with_optimistic_retry
from waku.eventsourcing.contracts.aggregate import IDecider  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.decider.repository import DeciderRepository  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.contracts.event import IEvent
from waku.messaging.contracts.request import RequestT, ResponseT
from waku.messaging.interfaces import IPublisher  # noqa: TC001  # Dishka needs runtime access
from waku.messaging.requests.handler import RequestHandler

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

__all__ = ['DeciderCommandHandler', 'DeciderVoidCommandHandler']

logger = logging.getLogger(__name__)

StateT = TypeVar('StateT', default=object)
CommandT = TypeVar('CommandT', default=object)
EventT = TypeVar('EventT', bound=IEvent, default=IEvent)


class DeciderCommandHandler(
    RequestHandler[RequestT, ResponseT],
    abc.ABC,
    Generic[RequestT, ResponseT, StateT, CommandT, EventT],
):
    max_attempts: ClassVar[int] = 3

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if 'max_attempts' in cls.__dict__ and cls.max_attempts < 1:
            msg = f'{cls.__name__}.max_attempts must be >= 1, got {cls.max_attempts}'
            raise ValueError(msg)

    def __init__(
        self,
        repository: DeciderRepository[StateT, CommandT, EventT],
        decider: IDecider[StateT, CommandT, EventT],
        publisher: IPublisher,
    ) -> None:
        self._repository = repository
        self._decider = decider
        self._publisher = publisher

    async def handle(self, request: RequestT, /) -> ResponseT:
        aggregate_id: str = self._aggregate_id(request)
        command: CommandT = self._to_command(request)
        idempotency_key: str | None = self._idempotency_key(request)
        logger.debug('Handling %s for %s', type(request).__name__, aggregate_id)

        async def _attempt() -> ResponseT:
            state, version = await self._repository.load(aggregate_id)

            events = self._decider.decide(command, state)
            for event in events:
                state = self._decider.evolve(state, event)

            new_version: int = await self._repository.save(
                aggregate_id,
                events,
                version,
                current_state=state,
                idempotency_key=idempotency_key,
            )

            for event in events:
                await self._publisher.publish(event)

            return self._to_response(state, new_version)

        return await execute_with_optimistic_retry(
            _attempt,
            max_attempts=self.max_attempts,
            request_name=type(request).__name__,
            aggregate_id=aggregate_id,
            attempt_context=self._create_attempt_context,
        )

    @abc.abstractmethod
    def _aggregate_id(self, request: RequestT) -> str: ...

    @abc.abstractmethod
    def _to_command(self, request: RequestT) -> CommandT: ...

    def _idempotency_key(self, request: RequestT) -> str | None:  # noqa: ARG002, PLR6301
        return None

    def _create_attempt_context(self) -> AbstractAsyncContextManager[Any]:  # noqa: PLR6301
        """Return a new context manager for a single retry attempt.

        Called once per attempt — must return a fresh instance each time.
        """
        return nullcontext()

    @abc.abstractmethod
    def _to_response(self, state: StateT, version: int) -> ResponseT: ...


class DeciderVoidCommandHandler(
    DeciderCommandHandler[RequestT, None, StateT, CommandT, EventT],
    abc.ABC,
    Generic[RequestT, StateT, CommandT, EventT],
):
    @override
    def _to_response(self, state: StateT, version: int) -> None:
        return None



from collections.abc import Sequence
from typing import Any, cast, overload

from dishka.exceptions import NoFactoryError
from typing_extensions import override

from waku.di import AsyncContainer
from waku.messaging import IPipelineBehavior
from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.contracts.request import IRequest, ResponseT
from waku.messaging.events.handler import IEventHandler
from waku.messaging.events.publish import EventPublisher
from waku.messaging.exceptions import RequestHandlerNotFound
from waku.messaging.interfaces import IMessageBus
from waku.messaging.pipeline import PipelineBehaviorWrapper
from waku.messaging.registry import MessageRegistry
from waku.messaging.requests.handler import RequestHandler


class MessageBus(IMessageBus):
    __slots__ = ('_container', '_event_publisher', '_registry')

    def __init__(
        self,
        container: AsyncContainer,
        event_publisher: EventPublisher,
        registry: MessageRegistry,
    ) -> None:
        self._container = container
        self._event_publisher = event_publisher
        self._registry = registry

    @overload
    async def invoke(self, request: IRequest[None], /) -> None: ...

    @overload
    async def invoke(self, request: IRequest[ResponseT], /) -> ResponseT: ...

    @override
    async def invoke(self, request: IRequest[Any], /) -> Any:
        request_type = type(request)
        handler = await self._resolve_request_handler(request_type)  # pyrefly: ignore[bad-argument-type]
        return await self._handle_request(handler, request)

    @override
    async def send(self, request: IRequest[Any], /) -> None:
        await self.invoke(request)

    @override
    async def publish(self, event: IEvent, /) -> None:
        event_type = type(event)
        handlers: Sequence[Any] = await self._resolve_event_handlers(event_type)  # pyrefly: ignore[bad-argument-type]
        await self._event_publisher(handlers, event)

    async def _resolve_request_handler(
        self,
        request_type: type[IRequest[ResponseT]],
    ) -> RequestHandler[IRequest[ResponseT], ResponseT]:
        if not self._registry.request_map.has_handler(request_type):
            raise RequestHandlerNotFound(request_type)

        handler_type = self._registry.request_map.get_handler_type(request_type)
        return cast('RequestHandler[IRequest[ResponseT], ResponseT]', await self._container.get(handler_type))

    async def _handle_request(
        self,
        handler: RequestHandler[IRequest[ResponseT], ResponseT],
        request: IRequest[ResponseT],
    ) -> ResponseT:
        request_type = type(request)
        behaviors = await self._resolve_behaviors(request_type)  # pyrefly: ignore[bad-argument-type]

        pipeline = PipelineBehaviorWrapper(behaviors).wrap(handler.handle)
        result = await pipeline(request)

        return cast('ResponseT', result)

    async def _resolve_behaviors(self, request_type: type[IRequest[Any]]) -> Sequence[IPipelineBehavior[Any, Any]]:
        try:
            global_behaviors = await self._container.get(Sequence[IPipelineBehavior[Any, Any]])
        except NoFactoryError:
            global_behaviors = ()

        if not self._registry.behavior_map.has_behaviors(request_type):
            return global_behaviors

        lookup_type = self._registry.behavior_map.get_lookup_type(request_type)
        scoped_behaviors = await self._container.get(Sequence[lookup_type])  # type: ignore[valid-type]

        return (*global_behaviors, *scoped_behaviors)

    async def _resolve_event_handlers(
        self,
        event_type: type[IEvent],
    ) -> Sequence[IEventHandler[EventT]]:
        if not self._registry.event_map.has_handlers(event_type):
            return ()

        handler_type = self._registry.event_map.get_handler_type(event_type)
        handlers = await self._container.get(Sequence[handler_type])  # type: ignore[valid-type]
        return cast('Sequence[IEventHandler[EventT]]', handlers)



from waku.eventsourcing.projection.lock.in_memory import InMemoryProjectionLock
from waku.eventsourcing.projection.lock.interfaces import IProjectionLock

__all__ = [
    'IProjectionLock',
    'InMemoryProjectionLock',
]



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self, TypeAlias

from typing_extensions import override

from waku.di import Provider, WithParents, many, object_, scoped
from waku.extensions import OnModuleConfigure, OnModuleRegistration
from waku.messaging.contracts.event import EventT
from waku.messaging.contracts.pipeline import IPipelineBehavior
from waku.messaging.contracts.request import RequestT
from waku.messaging.events.handler import EventHandler
from waku.messaging.events.publish import EventPublisher, SequentialEventPublisher
from waku.messaging.impl import MessageBus
from waku.messaging.interfaces import IMessageBus
from waku.messaging.registry import MessageRegistry
from waku.messaging.requests.handler import RequestHandler
from waku.modules import DynamicModule, ModuleMetadataRegistry, module

if TYPE_CHECKING:
    from waku.modules import ModuleMetadata, ModuleType

__all__ = [
    'MessagingConfig',
    'MessagingExtension',
    'MessagingModule',
]


_HandlerProviders: TypeAlias = tuple[Provider, ...]


@dataclass(frozen=True, slots=True, kw_only=True)
class MessagingConfig:
    """Configuration for the messaging extension.

    Attributes:
        event_publisher: The implementation class for publishing events. Defaults to `SequentialEventPublisher`.
        pipeline_behaviors: A sequence of pipeline behavior configurations that will be applied
            to the messaging pipeline. Behaviors are executed in the order they are defined.
            Defaults to an empty sequence.

    Example:
        ```python
        config = MessagingConfig(
            pipeline_behaviors=[
                LoggingBehavior,
                ValidationBehavior,
            ]
        )
        ```
    """

    event_publisher: type[EventPublisher] = SequentialEventPublisher
    pipeline_behaviors: Sequence[type[IPipelineBehavior[Any, Any]]] = ()


@module()
class MessagingModule:
    @classmethod
    def register(cls, config: MessagingConfig | None = None, /) -> DynamicModule:
        """Application-level module for MessageBus setup.

        Args:
            config: Configuration for the messaging extension.
        """
        config_ = config or MessagingConfig()
        return DynamicModule(
            parent_module=cls,
            providers=[
                *cls._create_bus_providers(config_),
                *cls._create_pipeline_behavior_providers(config_),
            ],
            extensions=[MessageRegistryAggregator()],
            is_global=True,
        )

    @staticmethod
    def _create_bus_providers(config: MessagingConfig) -> _HandlerProviders:
        return (
            scoped(WithParents[IMessageBus], MessageBus),  # ty:ignore[not-subscriptable]
            scoped(EventPublisher, config.event_publisher),
        )

    @staticmethod
    def _create_pipeline_behavior_providers(config: MessagingConfig) -> _HandlerProviders:
        if not config.pipeline_behaviors:
            return ()
        return (many(IPipelineBehavior[Any, Any], *config.pipeline_behaviors),)


class MessagingExtension(OnModuleConfigure):
    def __init__(self) -> None:
        self._registry = MessageRegistry()

    @override
    def on_module_configure(self, metadata: 'ModuleMetadata') -> None:
        pass

    def bind_request(
        self,
        request_type: type[RequestT],
        handler_type: type[RequestHandler[RequestT, Any]],
        *,
        behaviors: list[type[IPipelineBehavior[RequestT, Any]]] | None = None,
    ) -> Self:
        self._registry.request_map.bind(request_type, handler_type)
        if behaviors:
            self._registry.behavior_map.bind(request_type, behaviors)
        return self

    def bind_event(
        self,
        event_type: type[EventT],
        handler_types: list[type[EventHandler[EventT]]],
    ) -> Self:
        self._registry.event_map.bind(event_type, handler_types)
        return self

    @property
    def registry(self) -> MessageRegistry:
        return self._registry


class MessageRegistryAggregator(OnModuleRegistration):
    @override
    def on_module_registration(
        self,
        registry: ModuleMetadataRegistry,
        owning_module: 'ModuleType',
        context: Mapping[Any, Any] | None,
    ) -> None:
        aggregated = MessageRegistry()

        for module_type, ext in registry.find_extensions(MessagingExtension):
            aggregated.merge(ext.registry)
            for provider in ext.registry.handler_providers():
                registry.add_provider(module_type, provider)

        for provider in aggregated.collector_providers():
            registry.add_provider(owning_module, provider)

        aggregated.freeze()
        registry.add_provider(owning_module, object_(aggregated))



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from waku.validation._errors import ValidationError
    from waku.validation._extension import ValidationContext

__all__ = ['ValidationRule']


class ValidationRule(Protocol):
    def validate(self, context: ValidationContext) -> list[ValidationError]: ...



from __future__ import annotations

from typing import Any, Protocol, runtime_checkable

from typing_extensions import TypeVar

__all__ = [
    'IRequest',
    'RequestT',
    'ResponseT',
]


ResponseT = TypeVar('ResponseT', default=None, covariant=True)  # noqa: PLC0105


@runtime_checkable
class IRequest(Protocol[ResponseT]):
    """Marker interface for request-type objects (commands/queries).

    Example::

        @dataclass(frozen=True)
        class GetUserQuery(IRequest[UserDTO]):
            user_id: str


        @dataclass(frozen=True)
        class CreateOrderCommand(IRequest):  # void command, returns None
            order_id: str

    """

    __slots__ = ()


RequestT = TypeVar('RequestT', bound=IRequest[Any], contravariant=True)  # noqa: PLC0105



from __future__ import annotations

import functools
import typing

from typing_extensions import get_original_bases

from waku.messaging.contracts.request import IRequest

if typing.TYPE_CHECKING:
    from waku.messaging.contracts import ResponseT

__all__ = ['get_request_response_type']

_RESPONSE_T_DEFAULT: typing.Final[type[None]] = type(None)


def _is_request_origin(origin: type | None) -> bool:
    if origin is None:
        return False
    if origin is IRequest:
        return True
    return isinstance(origin, type) and issubclass(origin, IRequest)  # pyrefly: ignore[invalid-argument]


def _extract_response_from_bases(cls: type) -> type[ResponseT] | None:
    for base in get_original_bases(cls):
        origin = typing.get_origin(base)
        if not _is_request_origin(origin):
            if base is IRequest:
                return typing.cast('type[ResponseT]', _RESPONSE_T_DEFAULT)
            continue
        if args := typing.get_args(base):
            response_type = args[0]
            if isinstance(response_type, typing.TypeVar):
                continue
            return typing.cast('type[ResponseT]', response_type)
    return None


@functools.cache
def get_request_response_type(request_type: type[IRequest[ResponseT]]) -> type[ResponseT]:
    """Extract the response type from an IRequest implementation.

    Searches through the class hierarchy to find IRequest or its subclasses,
    supporting direct implementations and nested inheritance.
    Falls back to the TypeVar default when no explicit type argument is provided.

    Raises:
        TypeError: if response type cannot be extracted from the request type.
    """
    for cls in request_type.__mro__:
        if cls is object:
            msg = f'Could not extract response type from {request_type.__name__}'
            raise TypeError(msg)
        if response_type := _extract_response_from_bases(cls):
            return response_type  # type: ignore[return-value]

    msg = f'Could not extract response type from {request_type.__name__}'  # pragma: no cover
    raise TypeError(msg)  # pragma: no cover



from __future__ import annotations

from typing import Protocol, TypeVar, runtime_checkable

__all__ = [
    'EventT',
    'IEvent',
]


@runtime_checkable
class IEvent(Protocol):
    """Marker interface for event-type objects.

    This is a pure marker protocol with no required attributes or methods.
    Implement this protocol for domain events.

    Example::

        @dataclass(frozen=True)
        class OrderPlaced(IEvent):
            order_id: str
            customer_id: str

    """

    __slots__ = ()


EventT = TypeVar('EventT', bound=IEvent, contravariant=True)  # noqa: PLC0105



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dishka import Provider

    from waku.modules._metadata import ModuleMetadata, ModuleType

__all__ = ['ModuleMetadataRegistry']

_ExtT = TypeVar('_ExtT')


class ModuleMetadataRegistry:
    """Registry providing access to collected module metadata.

    Provides read access to all modules' metadata for aggregation purposes,
    with controlled write access through explicit methods.

    This class is used during the module registration phase to enable
    cross-module aggregation of providers.
    """

    __slots__ = ('_metadata_by_type', '_topological_order')

    def __init__(
        self,
        metadata_by_type: dict[ModuleType, ModuleMetadata],
        topological_order: tuple[ModuleType, ...],
    ) -> None:
        self._metadata_by_type = metadata_by_type
        self._topological_order = topological_order

    @property
    def modules(self) -> tuple[ModuleType, ...]:
        """All module types in topological order (dependencies first)."""
        return self._topological_order

    def get_metadata(self, module_type: ModuleType) -> ModuleMetadata:
        """Get metadata for a specific module."""
        return self._metadata_by_type[module_type]

    def find_extensions(self, protocol: type[_ExtT]) -> Iterator[tuple[ModuleType, _ExtT]]:
        """Find all extensions of a given type across all modules.

        Yields (module_type, extension) pairs in topological order.
        This is useful for aggregating data from extensions across modules.

        Args:
            protocol: The extension protocol/type to search for.

        Yields:
            Tuples of (module_type, extension) for each matching extension.
        """
        for module_type in self._topological_order:
            metadata = self._metadata_by_type[module_type]
            for ext in metadata.extensions:
                if isinstance(ext, protocol):
                    yield module_type, ext

    def add_provider(self, module_type: ModuleType, provider: Provider) -> None:
        """Add a provider to a module's metadata.

        This is the preferred way to add providers during registration hooks.
        The provider will become part of the owning module.

        Args:
            module_type: The module to add the provider to.
            provider: The provider specification to add.

        Raises:
            KeyError: If module_type is not in the registry.
        """
        self._metadata_by_type[module_type].providers.append(provider)



from __future__ import annotations

from collections import OrderedDict, defaultdict
from dataclasses import replace as _replace_dataclass
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Final, TypeAlias
from uuid import UUID

from waku.extensions import OnModuleRegistration
from waku.modules._metadata import ModuleCompiler, ModuleMetadata, ModuleType
from waku.modules._metadata_registry import ModuleMetadataRegistry
from waku.modules._module import Module
from waku.modules._registry import ModuleRegistry

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from waku import DynamicModule
    from waku.di import BaseProvider
    from waku.extensions import ApplicationExtension


__all__ = [
    'AdjacencyMatrix',
    'ModuleRegistryBuilder',
]


AdjacencyMatrix: TypeAlias = dict[UUID, OrderedDict[UUID, str]]


class ModuleRegistryBuilder:
    def __init__(
        self,
        root_module_type: ModuleType,
        compiler: ModuleCompiler | None = None,
        context: dict[Any, Any] | None = None,
        app_extensions: Sequence[ApplicationExtension] = (),
    ) -> None:
        self._compiler: Final = compiler or ModuleCompiler()
        self._root_module_type: Final = root_module_type
        self._context: Final = context
        self._app_extensions: Final = app_extensions
        self._modules: dict[UUID, Module] = {}
        self._providers: list[BaseProvider] = []

        self._metadata_cache: dict[ModuleType | DynamicModule, tuple[ModuleType, ModuleMetadata]] = {}

    def build(self) -> ModuleRegistry:
        modules, adjacency = self._collect_modules()
        self._execute_registration_hooks(modules)
        root_module = self._register_modules(modules)
        return self._build_registry(root_module, adjacency)

    def _collect_modules(self) -> tuple[list[tuple[ModuleType, ModuleMetadata]], AdjacencyMatrix]:
        visited: set[UUID] = set()
        post_order: list[tuple[ModuleType, ModuleMetadata]] = []
        adjacency: AdjacencyMatrix = defaultdict(OrderedDict)
        self._collect_modules_recursive(self._root_module_type, visited, post_order, adjacency)
        return post_order, adjacency

    def _collect_modules_recursive(
        self,
        current_type: ModuleType | DynamicModule,
        visited: set[UUID],
        post_order: list[tuple[ModuleType, ModuleMetadata]],
        adjacency: AdjacencyMatrix,
    ) -> None:
        type_, metadata = self._get_metadata(current_type)
        if metadata.id in visited:
            return

        adjacency[metadata.id][metadata.id] = type_.__name__

        for imported in metadata.imports:
            imported_type, imported_metadata = self._get_metadata(imported)
            adjacency[metadata.id][imported_metadata.id] = imported_type.__name__
            if imported_metadata.id not in visited:
                self._collect_modules_recursive(imported, visited, post_order, adjacency)

        # Isolate build-phase mutations (add_provider, is_global) from cached originals
        post_order.append((type_, self._copy_metadata(metadata)))
        visited.add(metadata.id)

    def _execute_registration_hooks(
        self,
        modules: list[tuple[ModuleType, ModuleMetadata]],
    ) -> None:
        metadata_by_type = dict(modules)
        topological_order = tuple(mod_type for mod_type, _ in modules)
        registry = ModuleMetadataRegistry(
            metadata_by_type=metadata_by_type,
            topological_order=topological_order,
        )

        read_only_context: Mapping[Any, Any] | None = MappingProxyType(self._context) if self._context else None

        for ext in self._app_extensions:
            if isinstance(ext, OnModuleRegistration):
                ext.on_module_registration(registry, self._root_module_type, read_only_context)

        for module_type, metadata in modules:
            for ext in metadata.extensions:
                if isinstance(ext, OnModuleRegistration):
                    ext.on_module_registration(registry, module_type, read_only_context)

    def _register_modules(self, post_order: list[tuple[ModuleType, ModuleMetadata]]) -> Module:
        for type_, metadata in post_order:
            if metadata.id in self._modules:
                continue

            if type_ is self._root_module_type:
                metadata.is_global = True

            module = Module(type_, metadata)

            self._modules[module.id] = module
            self._providers.append(module.create_provider())

        _, root_metadata = self._get_metadata(self._root_module_type)
        return self._modules[root_metadata.id]

    def _get_metadata(self, module_type: ModuleType | DynamicModule) -> tuple[ModuleType, ModuleMetadata]:
        if module_type not in self._metadata_cache:
            self._metadata_cache[module_type] = self._compiler.extract_metadata(module_type)
        return self._metadata_cache[module_type]

    @staticmethod
    def _copy_metadata(metadata: ModuleMetadata) -> ModuleMetadata:
        return _replace_dataclass(
            metadata,
            providers=list(metadata.providers),
            imports=list(metadata.imports),
            exports=list(metadata.exports),
            extensions=list(metadata.extensions),
        )

    def _build_registry(self, root_module: Module, adjacency: AdjacencyMatrix) -> ModuleRegistry:
        return ModuleRegistry(
            compiler=self._compiler,
            modules=self._modules,
            providers=self._providers,
            root_module=root_module,
            adjacency=adjacency,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlalchemy import (  # Dishka needs runtime access
    Table,
    select,
)
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.ext.asyncio import AsyncSession  # noqa: TC002  # Dishka needs runtime access

from waku.eventsourcing.projection.checkpoint import Checkpoint
from waku.eventsourcing.projection.interfaces import ICheckpointStore

if TYPE_CHECKING:
    from collections.abc import Callable

__all__ = [
    'SqlAlchemyCheckpointStore',
    'make_sqlalchemy_checkpoint_store',
]


class SqlAlchemyCheckpointStore(ICheckpointStore):
    def __init__(self, session: AsyncSession, checkpoints_table: Table) -> None:
        self._session = session
        self._checkpoints = checkpoints_table

    async def load(self, projection_name: str, /) -> Checkpoint | None:
        query = select(self._checkpoints).where(self._checkpoints.c.projection_name == projection_name)
        result = await self._session.execute(query)
        row: Any = result.one_or_none()
        if row is None:
            return None
        return Checkpoint(
            projection_name=row.projection_name,
            position=row.position,
            updated_at=row.updated_at,
        )

    async def save(self, checkpoint: Checkpoint, /) -> None:
        stmt = pg_insert(self._checkpoints).values(
            projection_name=checkpoint.projection_name,
            position=checkpoint.position,
            updated_at=checkpoint.updated_at,
        )
        stmt = stmt.on_conflict_do_update(
            index_elements=['projection_name'],
            set_={
                'position': stmt.excluded.position,
                'updated_at': stmt.excluded.updated_at,
            },
        )
        await self._session.execute(stmt)
        await self._session.flush()


def make_sqlalchemy_checkpoint_store(
    checkpoints_table: Table,
) -> Callable[..., SqlAlchemyCheckpointStore]:
    def factory(session: AsyncSession) -> SqlAlchemyCheckpointStore:
        return SqlAlchemyCheckpointStore(session, checkpoints_table)

    return factory



from __future__ import annotations

import contextlib
import logging
import uuid
from typing import TYPE_CHECKING

import anyio
from sqlalchemy import text

from waku.eventsourcing.projection.lock.interfaces import IProjectionLock

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from sqlalchemy.ext.asyncio import AsyncEngine

    from waku.eventsourcing.projection.config import LeaseConfig

__all__ = ['PostgresLeaseProjectionLock']

logger = logging.getLogger(__name__)

_UPSERT_SQL = text("""\
INSERT INTO es_projection_leases (projection_name, holder_id, expires_at)
VALUES (:name, :holder, now() + make_interval(secs => :ttl))
ON CONFLICT (projection_name) DO UPDATE
SET holder_id = EXCLUDED.holder_id,
    acquired_at = now(),
    renewed_at = now(),
    expires_at = now() + make_interval(secs => :ttl)
WHERE es_projection_leases.expires_at < now()
RETURNING projection_name;
""")

_RENEW_SQL = text("""\
UPDATE es_projection_leases
SET renewed_at = now(), expires_at = now() + make_interval(secs => :ttl)
WHERE projection_name = :name AND holder_id = :holder;
""")

_RELEASE_SQL = text("""\
DELETE FROM es_projection_leases
WHERE projection_name = :name AND holder_id = :holder;
""")


class PostgresLeaseProjectionLock(IProjectionLock):
    """Production lease-based projection lock backed by PostgreSQL."""

    def __init__(self, engine: AsyncEngine, config: LeaseConfig) -> None:
        self._engine = engine
        self._config = config
        self._holder_id = str(uuid.uuid4())

    @contextlib.asynccontextmanager
    async def acquire(self, projection_name: str) -> AsyncGenerator[bool]:
        async with self._engine.connect() as conn:
            await conn.execution_options(isolation_level='AUTOCOMMIT')
            result = await conn.execute(
                _UPSERT_SQL,
                {'name': projection_name, 'holder': self._holder_id, 'ttl': self._config.ttl_seconds},
            )
            row = result.fetchone()

        if row is None:
            yield False
            return

        logger.debug('Lease acquired for %s by %s', projection_name, self._holder_id)

        try:
            async with anyio.create_task_group() as tg:
                tg.start_soon(self._heartbeat, projection_name, tg.cancel_scope)
                try:
                    yield True
                finally:
                    tg.cancel_scope.cancel()
        finally:
            await self._release(projection_name)

    async def _heartbeat(self, projection_name: str, cancel_scope: anyio.CancelScope) -> None:
        while not cancel_scope.cancel_called:
            await anyio.sleep(self._config.renew_interval_seconds)

            async with self._engine.connect() as conn:
                await conn.execution_options(isolation_level='AUTOCOMMIT')
                result = await conn.execute(
                    _RENEW_SQL,
                    {'name': projection_name, 'holder': self._holder_id, 'ttl': self._config.ttl_seconds},
                )

            if result.rowcount == 0:
                logger.warning('Lease for %s was stolen from holder %s', projection_name, self._holder_id)
                cancel_scope.cancel()
                return

            logger.debug('Lease renewed for %s by %s', projection_name, self._holder_id)

    async def _release(self, projection_name: str) -> None:
        try:
            async with self._engine.connect() as conn:
                await conn.execution_options(isolation_level='AUTOCOMMIT')
                await conn.execute(
                    _RELEASE_SQL,
                    {'name': projection_name, 'holder': self._holder_id},
                )
            logger.debug('Lease released for %s by %s', projection_name, self._holder_id)
        except Exception:
            logger.warning('Failed to release lease for %s by %s', projection_name, self._holder_id, exc_info=True)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from waku.eventsourcing.exceptions import SnapshotTypeMismatchError
from waku.eventsourcing.snapshot.interfaces import Snapshot
from waku.eventsourcing.snapshot.migration import migrate_snapshot_or_discard

if TYPE_CHECKING:
    from waku.eventsourcing.contracts.stream import StreamId
    from waku.eventsourcing.snapshot.interfaces import ISnapshotStore
    from waku.eventsourcing.snapshot.registry import SnapshotConfig

logger = logging.getLogger(__name__)


class SnapshotManager:
    __slots__ = (
        '_config',
        '_last_snapshot_versions',
        '_state_type_name',
        '_store',
    )

    def __init__(
        self,
        store: ISnapshotStore,
        config: SnapshotConfig,
        state_type_name: str,
    ) -> None:
        self._store = store
        self._config = config
        self._state_type_name = state_type_name
        self._last_snapshot_versions: dict[str, int] = {}

    async def load_snapshot(self, stream_id: StreamId, aggregate_id: str) -> Snapshot | None:
        snapshot = await self._store.load(stream_id)

        if snapshot is None:
            self._last_snapshot_versions[aggregate_id] = -1
            return None

        if snapshot.state_type != self._state_type_name:
            raise SnapshotTypeMismatchError(stream_id, self._state_type_name, snapshot.state_type)

        if snapshot.schema_version != self._config.schema_version:
            snapshot = migrate_snapshot_or_discard(
                self._config.migration_chain,
                snapshot,
                self._config.schema_version,
                stream_id,
            )
            if snapshot is None:
                self._last_snapshot_versions[aggregate_id] = -1
                return None

        self._last_snapshot_versions[aggregate_id] = snapshot.version
        return snapshot

    def should_save(self, aggregate_id: str, new_version: int) -> bool:
        last_snapshot_version = self._last_snapshot_versions.get(aggregate_id, -1)
        events_since_snapshot = new_version - last_snapshot_version
        return self._config.strategy.should_snapshot(new_version, events_since_snapshot)

    async def save_snapshot(
        self,
        stream_id: StreamId,
        aggregate_id: str,
        state_data: dict[str, Any],
        version: int,
    ) -> None:
        snapshot = Snapshot(
            stream_id=stream_id,
            state=state_data,
            version=version,
            state_type=self._state_type_name,
            schema_version=self._config.schema_version,
        )
        try:
            await self._store.save(snapshot)
        except Exception:
            logger.warning(
                'Failed to save snapshot for stream %s at version %d, continuing without snapshot',
                stream_id,
                version,
                exc_info=True,
            )
            return
        self._last_snapshot_versions[aggregate_id] = version



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from waku.eventsourcing.projection.interfaces import ErrorPolicy

if TYPE_CHECKING:
    from waku.eventsourcing.projection.interfaces import ICatchUpProjection

__all__ = ['CatchUpProjectionBinding']


@dataclass(frozen=True, slots=True, kw_only=True)
class CatchUpProjectionBinding:
    projection: type[ICatchUpProjection]
    error_policy: ErrorPolicy = ErrorPolicy.STOP
    max_retry_attempts: int = 0
    base_retry_delay_seconds: float = 10.0
    max_retry_delay_seconds: float = 300.0
    batch_size: int = 100
    event_type_names: tuple[str, ...] | None = None
    gap_detection_enabled: bool = False
    gap_timeout_seconds: float = 10.0



from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.contracts.pipeline import IPipelineBehavior, NextHandlerType
from waku.messaging.contracts.request import IRequest, RequestT, ResponseT

__all__ = [
    'EventT',
    'IEvent',
    'IPipelineBehavior',
    'IRequest',
    'NextHandlerType',
    'RequestT',
    'ResponseT',
]



from __future__ import annotations

from typing import TYPE_CHECKING, Generic

from waku.messaging.contracts.request import RequestT, ResponseT

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.messaging.contracts.pipeline import IPipelineBehavior, NextHandlerType


class PipelineBehaviorWrapper(Generic[RequestT, ResponseT]):
    """Composes pipeline behaviors into a processing chain."""

    __slots__ = ('_behaviors',)

    def __init__(self, behaviors: Sequence[IPipelineBehavior[RequestT, ResponseT]]) -> None:
        self._behaviors = tuple(behaviors)

    def wrap(self, handle: NextHandlerType[RequestT, ResponseT]) -> NextHandlerType[RequestT, ResponseT]:
        """Create a pipeline that wraps the handler function with behaviors.

        Pipeline behaviors are executed in the order they are provided.

        Args:
            handle: The handler function to wrap with behaviors

        Returns:
            A function that executes the entire pipeline
        """
        if not self._behaviors:
            return handle

        behaviors = self._behaviors

        async def pipeline(request: RequestT) -> ResponseT:
            async def execute(req: RequestT, idx: int) -> ResponseT:
                if idx >= len(behaviors):
                    return await handle(req)
                return await behaviors[idx].handle(req, next_handler=lambda r: execute(r, idx + 1))

            return await execute(request, 0)

        return pipeline



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence  # noqa: TC003  # Dishka needs runtime access
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, Protocol, assert_never

from sqlalchemy import (  # Dishka needs runtime access
    func as sa_func,
    select,
)
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession  # noqa: TC002  # Dishka needs runtime access

from waku.eventsourcing.contracts.event import EventMetadata, IMetadataEnricher, StoredEvent
from waku.eventsourcing.contracts.stream import StreamId, StreamPosition
from waku.eventsourcing.exceptions import (
    ConcurrencyConflictError,
    DuplicateIdempotencyKeyError,
    PartialDuplicateAppendError,
    StreamDeletedError,
    StreamNotFoundError,
)
from waku.eventsourcing.projection.interfaces import IProjection  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.serialization.interfaces import IEventSerializer  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.serialization.registry import EventTypeRegistry  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.store._shared import enrich_metadata
from waku.eventsourcing.store._version_check import check_expected_version
from waku.eventsourcing.store.interfaces import IEventStore
from waku.eventsourcing.store.sqlalchemy._serialization import row_to_stored_event, serialize_metadata
from waku.eventsourcing.store.sqlalchemy.tables import (  # Dishka needs runtime access
    IDEMPOTENCY_KEY_CONSTRAINT,
    EventStoreTables,
)
from waku.eventsourcing.upcasting.chain import UpcasterChain  # noqa: TC001  # Dishka needs runtime access

if TYPE_CHECKING:
    from waku.eventsourcing.contracts.event import EventEnvelope
    from waku.eventsourcing.contracts.stream import ExpectedVersion

__all__ = [
    'SqlAlchemyEventStore',
    'make_sqlalchemy_event_store',
]

logger = logging.getLogger(__name__)


class SqlAlchemyEventStoreFactory(Protocol):
    def __call__(
        self,
        session: AsyncSession,
        serializer: IEventSerializer,
        registry: EventTypeRegistry,
        upcaster_chain: UpcasterChain,
        projections: Sequence[IProjection] = (),
        enrichers: Sequence[IMetadataEnricher] = (),
    ) -> SqlAlchemyEventStore: ...


class SqlAlchemyEventStore(IEventStore):
    def __init__(
        self,
        session: AsyncSession,
        serializer: IEventSerializer,
        registry: EventTypeRegistry,
        tables: EventStoreTables,
        upcaster_chain: UpcasterChain,
        projections: Sequence[IProjection] = (),
        enrichers: Sequence[IMetadataEnricher] = (),
    ) -> None:
        self._session = session
        self._serializer = serializer
        self._registry = registry
        self._streams = tables.streams
        self._events = tables.events
        self._upcaster_chain = upcaster_chain
        self._projections = projections
        self._enrichers = enrichers

    @property
    def _not_deleted(self) -> Any:
        return self._streams.c.deleted_at.is_(None)

    async def read_stream(
        self,
        stream_id: StreamId,
        /,
        *,
        start: int | StreamPosition = StreamPosition.START,
        count: int | None = None,
    ) -> list[StoredEvent]:
        key = str(stream_id)

        if count == 0:
            await self._ensure_stream_exists(stream_id)
            return []

        if start is StreamPosition.END:
            return await self._read_stream_end(stream_id, key)

        match start:
            case StreamPosition.START:
                offset = 0
            case int() as offset:
                pass
            case _:  # pragma: no cover
                assert_never(start)

        query = (
            select(self._events)
            .where(self._events.c.stream_id == key)
            .where(self._events.c.position >= offset)
            .order_by(self._events.c.position)
        )
        if count is not None:
            query = query.limit(count)

        result = await self._session.execute(query)
        rows = result.fetchall()

        if not rows:
            await self._ensure_stream_exists(stream_id)

        return [
            row_to_stored_event(
                row, registry=self._registry, upcaster_chain=self._upcaster_chain, serializer=self._serializer
            )
            for row in rows
        ]

    async def _read_stream_end(self, stream_id: StreamId, key: str) -> list[StoredEvent]:
        query = (
            select(self._events)
            .where(self._events.c.stream_id == key)
            .order_by(self._events.c.position.desc())
            .limit(1)
        )
        result = await self._session.execute(query)
        row: Any = result.one_or_none()

        if row is None:
            raise StreamNotFoundError(stream_id)

        return [
            row_to_stored_event(
                row, registry=self._registry, upcaster_chain=self._upcaster_chain, serializer=self._serializer
            )
        ]

    async def _ensure_stream_exists(self, stream_id: StreamId) -> None:
        stream_row = await self._get_stream(stream_id)
        if stream_row is None:
            raise StreamNotFoundError(stream_id)

    async def read_all(
        self,
        *,
        after_position: int = -1,
        count: int | None = None,
        event_types: Sequence[str] | None = None,
    ) -> list[StoredEvent]:
        query = (
            select(self._events)
            .join(self._streams, self._events.c.stream_id == self._streams.c.stream_id)
            .where(self._events.c.global_position > after_position)
            .where(self._not_deleted)
            .order_by(self._events.c.global_position)
        )
        if event_types:
            query = query.where(self._events.c.event_type.in_(event_types))
        if count is not None:
            query = query.limit(count)

        result = await self._session.execute(query)
        rows = result.fetchall()
        return [
            row_to_stored_event(
                row, registry=self._registry, upcaster_chain=self._upcaster_chain, serializer=self._serializer
            )
            for row in rows
        ]

    async def stream_exists(self, stream_id: StreamId, /) -> bool:
        key = str(stream_id)
        query = select(self._streams.c.stream_id).where(
            self._streams.c.stream_id == key,
            self._not_deleted,
        )
        result = await self._session.execute(query)
        return result.scalar_one_or_none() is not None

    async def global_head_position(self) -> int:
        query = select(sa_func.coalesce(sa_func.max(self._events.c.global_position), -1))
        result = await self._session.execute(query)
        return int(result.scalar_one())

    async def read_positions(
        self,
        *,
        after_position: int,
        up_to_position: int,
    ) -> list[int]:
        query = (
            select(self._events.c.global_position)
            .join(self._streams, self._events.c.stream_id == self._streams.c.stream_id)
            .where(self._events.c.global_position > after_position)
            .where(self._events.c.global_position <= up_to_position)
            .where(self._not_deleted)
            .order_by(self._events.c.global_position)
        )
        result = await self._session.execute(query)
        return [row[0] for row in result.fetchall()]

    async def delete_stream(self, stream_id: StreamId, /) -> None:
        stream_row = await self._get_stream(stream_id)
        if stream_row is None:
            raise StreamNotFoundError(stream_id)
        if stream_row.deleted_at is not None:
            return
        await self._session.execute(
            self._streams.update().where(self._streams.c.stream_id == str(stream_id)).values(deleted_at=sa_func.now())
        )

    async def append_to_stream(
        self,
        stream_id: StreamId,
        /,
        events: Sequence[EventEnvelope],
        *,
        expected_version: ExpectedVersion,
    ) -> int:
        if not events:
            return await self._resolve_current_version(stream_id, expected_version)

        dedup_version = await self._check_idempotency(stream_id, events)
        if dedup_version is not None:
            return dedup_version

        current_version = await self._resolve_current_version(stream_id, expected_version)
        new_version = current_version + len(events)

        try:
            async with self._session.begin_nested():
                await self._ensure_stream_row(stream_id)
                await self._update_stream_version(stream_id, current_version, new_version)
                stored_events = await self._insert_events(stream_id, events, start_position=current_version + 1)
        except IntegrityError as exc:
            if IDEMPOTENCY_KEY_CONSTRAINT in str(exc):
                logger.warning(
                    'Idempotency race condition on stream %s: duplicate key caught by DB constraint',
                    stream_id,
                )
                dedup_version = await self._check_idempotency(stream_id, events)
                if dedup_version is not None:
                    return dedup_version
                logger.exception(  # pragma: no cover
                    'Idempotency re-check returned no match after IntegrityError on stream %s — '
                    'this should not happen under normal conditions',
                    stream_id,
                )
                raise DuplicateIdempotencyKeyError(
                    stream_id,
                    reason='conflict with existing keys',
                ) from exc  # pragma: no cover
            raise  # pragma: no cover

        for projection in self._projections:
            await projection.project(stored_events)

        return new_version

    async def _check_idempotency(
        self,
        stream_id: StreamId,
        events: Sequence[EventEnvelope],
    ) -> int | None:
        keys = [e.idempotency_key for e in events]
        unique_keys = set(keys)
        if len(unique_keys) != len(keys):
            raise DuplicateIdempotencyKeyError(stream_id, reason='duplicate keys within batch')

        key = str(stream_id)
        query = select(self._events.c.idempotency_key).where(
            self._events.c.stream_id == key,
            self._events.c.idempotency_key.in_(keys),
        )
        result = await self._session.execute(query)
        existing_keys = {row[0] for row in result.fetchall()}

        if not existing_keys:
            return None

        if existing_keys == unique_keys:
            stream_row = await self._get_stream(stream_id)
            if stream_row.deleted_at is not None:
                raise StreamDeletedError(stream_id)
            return int(stream_row.version)  # stream must exist if events with these keys exist

        raise PartialDuplicateAppendError(stream_id, len(existing_keys), len(keys))

    async def _resolve_current_version(
        self,
        stream_id: StreamId,
        expected_version: ExpectedVersion,
    ) -> int:
        stream_row = await self._get_stream(stream_id)
        if stream_row is not None and stream_row.deleted_at is not None:
            raise StreamDeletedError(stream_id)
        current_version = stream_row.version if stream_row is not None else -1
        check_expected_version(stream_id, expected_version, current_version, exists=stream_row is not None)
        return current_version

    async def _ensure_stream_row(self, stream_id: StreamId) -> None:
        key = str(stream_id)
        await self._session.execute(
            pg_insert(self._streams)
            .values(
                stream_id=key,
                stream_type=stream_id.stream_type,
                version=-1,
            )
            .on_conflict_do_nothing(index_elements=['stream_id'])
        )

    async def _update_stream_version(
        self,
        stream_id: StreamId,
        expected_version: int,
        new_version: int,
    ) -> None:
        key = str(stream_id)
        result = await self._session.execute(
            self._streams
            .update()
            .where(
                self._streams.c.stream_id == key,
                self._streams.c.version == expected_version,
            )
            .values(
                version=new_version,
                updated_at=sa_func.now(),
            )
        )
        if result.rowcount != 1:  # type: ignore[attr-defined]  # pragma: no cover
            raise ConcurrencyConflictError(stream_id, expected_version, new_version)

    async def _insert_events(
        self,
        stream_id: StreamId,
        events: Sequence[EventEnvelope],
        *,
        start_position: int,
    ) -> list[StoredEvent]:
        key = str(stream_id)
        rows: list[dict[str, Any]] = []
        envelopes_data: list[tuple[uuid.UUID, str, datetime, object, EventMetadata]] = []

        position = start_position
        for envelope in events:
            event_id = uuid.uuid4()
            now = datetime.now(UTC)
            event_type = self._registry.get_name(type(envelope.domain_event))  # pyrefly: ignore[bad-argument-type]
            metadata = enrich_metadata(envelope.metadata, self._enrichers)

            rows.append({
                'event_id': event_id,
                'stream_id': key,
                'event_type': event_type,
                'position': position,
                'data': self._serializer.serialize(envelope.domain_event),
                'metadata': serialize_metadata(metadata),
                'timestamp': now,
                'schema_version': self._registry.get_version(
                    type(envelope.domain_event)  # pyrefly: ignore[bad-argument-type]
                ),
                'idempotency_key': envelope.idempotency_key,
            })
            envelopes_data.append((event_id, event_type, now, envelope.domain_event, metadata))
            position += 1

        result = await self._session.execute(
            self._events.insert().values(rows).returning(self._events.c.global_position)
        )
        global_positions = [row[0] for row in result.fetchall()]

        return [
            StoredEvent(
                event_id=envelopes_data[i][0],
                stream_id=stream_id,
                event_type=envelopes_data[i][1],
                position=rows[i]['position'],
                global_position=global_positions[i],
                timestamp=envelopes_data[i][2],
                data=envelopes_data[i][3],
                metadata=envelopes_data[i][4],
                idempotency_key=events[i].idempotency_key,
                schema_version=rows[i]['schema_version'],
            )
            for i in range(len(events))
        ]

    async def _get_stream(self, stream_id: StreamId, /) -> Any:
        query = select(self._streams).where(self._streams.c.stream_id == str(stream_id))
        result = await self._session.execute(query)
        return result.one_or_none()


def make_sqlalchemy_event_store(tables: EventStoreTables) -> SqlAlchemyEventStoreFactory:
    def factory(
        session: AsyncSession,
        serializer: IEventSerializer,
        registry: EventTypeRegistry,
        upcaster_chain: UpcasterChain,
        projections: Sequence[IProjection] = (),
        enrichers: Sequence[IMetadataEnricher] = (),
    ) -> SqlAlchemyEventStore:
        return SqlAlchemyEventStore(session, serializer, registry, tables, upcaster_chain, projections, enrichers)

    return factory



from __future__ import annotations

from sqlalchemy import Column, MetaData, Table, Text, func
from sqlalchemy.dialects.postgresql import TIMESTAMP

__all__ = ['bind_lease_tables']

_internal_metadata = MetaData()

es_projection_leases_table = Table(
    'es_projection_leases',
    _internal_metadata,
    Column('projection_name', Text, primary_key=True),
    Column('holder_id', Text, nullable=False),
    Column('acquired_at', TIMESTAMP(timezone=True), server_default=func.now()),
    Column('renewed_at', TIMESTAMP(timezone=True), server_default=func.now()),
    Column('expires_at', TIMESTAMP(timezone=True), nullable=False),
)


def bind_lease_tables(metadata: MetaData) -> Table:
    if es_projection_leases_table.name in metadata.tables:
        return metadata.tables[es_projection_leases_table.name]
    return es_projection_leases_table.to_metadata(metadata)



from __future__ import annotations

import random

__all__ = [
    'AdaptiveInterval',
    'calculate_backoff_with_jitter',
]


class AdaptiveInterval:
    """Fast when busy, slow when idle."""

    __slots__ = ('_current', '_jitter_factor', '_max', '_min', '_step')

    def __init__(
        self,
        min_seconds: float,
        max_seconds: float,
        step_seconds: float,
        jitter_factor: float = 0.1,
    ) -> None:
        self._min = min_seconds
        self._max = max_seconds
        self._step = step_seconds
        self._jitter_factor = jitter_factor
        self._current = min_seconds

    @property
    def current(self) -> float:
        return self._current

    def current_with_jitter(self) -> float:
        return self._current * random.uniform(1 - self._jitter_factor, 1 + self._jitter_factor)  # noqa: S311

    def on_work_done(self) -> None:
        self._current = self._min

    def on_idle(self) -> None:
        self._current = min(self._current + self._step, self._max)


def calculate_backoff_with_jitter(
    attempt: int,
    base_delay_seconds: float,
    max_delay_seconds: float,
) -> float:
    """Full jitter: random(0, min(base * 2^attempt, max_delay))."""
    max_delay = min(base_delay_seconds * (2**attempt), max_delay_seconds)
    return random.uniform(0, max_delay)  # noqa: S311



from __future__ import annotations

from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import Any, Generic, Self, TypeAlias

from waku.messaging.contracts.pipeline import IPipelineBehavior
from waku.messaging.contracts.request import IRequest, RequestT, ResponseT
from waku.messaging.exceptions import MapFrozenError, PipelineBehaviorAlreadyRegistered
from waku.messaging.utils import get_request_response_type

__all__ = [
    'PipelineBehaviorMap',
    'PipelineBehaviorMapEntry',
    'PipelineBehaviorMapRegistry',
]


@dataclass(frozen=True, slots=True)
class PipelineBehaviorMapEntry(Generic[RequestT, ResponseT]):
    request_type: type[IRequest[ResponseT]]
    di_lookup_type: type[IPipelineBehavior[RequestT, ResponseT]]
    behavior_types: list[type[IPipelineBehavior[RequestT, ResponseT]]] = field(default_factory=list)

    @classmethod
    def for_request(cls, request_type: type[IRequest[ResponseT]]) -> Self:
        response_type = get_request_response_type(request_type)
        di_lookup_type = IPipelineBehavior[request_type, response_type]  # type: ignore[valid-type]
        return cls(request_type=request_type, di_lookup_type=di_lookup_type)  # type: ignore[type-abstract]

    def add(self, behavior_type: type[IPipelineBehavior[RequestT, ResponseT]]) -> None:
        if behavior_type in self.behavior_types:
            raise PipelineBehaviorAlreadyRegistered(self.request_type, behavior_type)
        self.behavior_types.append(behavior_type)


PipelineBehaviorMapRegistry: TypeAlias = MutableMapping[type[RequestT], PipelineBehaviorMapEntry[RequestT, ResponseT]]


class PipelineBehaviorMap:
    def __init__(self) -> None:
        self._registry: PipelineBehaviorMapRegistry[Any, Any] = {}
        self._frozen = False

    def freeze(self) -> None:
        self._frozen = True

    @property
    def is_frozen(self) -> bool:
        return self._frozen

    def bind(
        self,
        request_type: type[RequestT],
        behavior_types: list[type[IPipelineBehavior[RequestT, ResponseT]]],
    ) -> Self:
        if self._frozen:
            raise MapFrozenError
        if request_type not in self._registry:
            self._registry[request_type] = PipelineBehaviorMapEntry.for_request(request_type)

        entry = self._registry[request_type]
        for behavior_type in behavior_types:
            entry.add(behavior_type)
        return self

    def merge(self, other: PipelineBehaviorMap) -> Self:
        if self._frozen:
            raise MapFrozenError
        for request_type, entry in other._registry.items():
            self.bind(request_type, entry.behavior_types)
        return self

    @property
    def registry(self) -> PipelineBehaviorMapRegistry[Any, Any]:
        return self._registry

    def has_behaviors(self, request_type: type[RequestT]) -> bool:
        return request_type in self._registry and len(self._registry[request_type].behavior_types) > 0

    def get_lookup_type(self, request_type: type[RequestT]) -> type[IPipelineBehavior[Any, Any]]:
        return self._registry[request_type].di_lookup_type

    def __bool__(self) -> bool:
        return bool(self._registry)



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from waku.eventsourcing.projection.lock.interfaces import IProjectionLock

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

__all__ = ['InMemoryProjectionLock']


class InMemoryProjectionLock(IProjectionLock):
    """Always acquires in single-process. Tracks held locks for testing."""

    def __init__(self) -> None:
        self._held: set[str] = set()

    @contextlib.asynccontextmanager
    async def acquire(self, projection_name: str) -> AsyncGenerator[bool]:
        if projection_name in self._held:
            yield False
            return

        self._held.add(projection_name)
        try:
            yield True
        finally:
            self._held.discard(projection_name)



# mypy: disable-error-code="type-abstract"
from __future__ import annotations

from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Self

from waku.extensions import (
    AfterApplicationInit,
    ExtensionRegistry,
    OnApplicationInit,
    OnApplicationShutdown,
    OnModuleDestroy,
    OnModuleInit,
)
from waku.lifespan import LifespanFunc, LifespanWrapper

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from types import TracebackType

    from waku.di import AsyncContainer
    from waku.modules import Module, ModuleRegistry

__all__ = ['WakuApplication']


class WakuApplication:
    __slots__ = (
        '_container',
        '_exit_stack',
        '_extension_registry',
        '_initialized',
        '_lifespan',
        '_registry',
    )

    def __init__(
        self,
        *,
        container: AsyncContainer,
        registry: ModuleRegistry,
        lifespan: Sequence[LifespanFunc | LifespanWrapper],
        extension_registry: ExtensionRegistry,
    ) -> None:
        self._container = container
        self._registry = registry
        self._lifespan = tuple(
            LifespanWrapper(lifespan_func) if not isinstance(lifespan_func, LifespanWrapper) else lifespan_func
            for lifespan_func in lifespan
        )
        self._extension_registry = extension_registry

        self._exit_stack = AsyncExitStack()
        self._initialized = False

    async def initialize(self) -> None:
        if self._initialized:
            return
        await self._call_on_init_extensions()
        self._initialized = True
        await self._call_after_init_extensions()

    async def close(self) -> None:
        if not self._initialized:
            return
        await self._call_on_shutdown_extensions()
        self._initialized = False

    @property
    def container(self) -> AsyncContainer:
        return self._container

    @property
    def registry(self) -> ModuleRegistry:
        return self._registry

    async def __aenter__(self) -> Self:
        await self.initialize()
        await self._exit_stack.__aenter__()
        for lifespan_wrapper in self._lifespan:
            await self._exit_stack.enter_async_context(lifespan_wrapper.lifespan(self))
        await self._exit_stack.enter_async_context(self._container)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def _call_on_init_extensions(self) -> None:
        # Call module OnModuleInit hooks sequentially in topological order (dependencies first)
        for module in self._get_modules_for_triggering_extensions():
            for module_ext in self._extension_registry.get_module_extensions(module.target, OnModuleInit):
                await module_ext.on_module_init(module)

        for app_ext in self._extension_registry.get_application_extensions(OnApplicationInit):
            await app_ext.on_app_init(self)

    async def _call_after_init_extensions(self) -> None:
        for extension in self._extension_registry.get_application_extensions(AfterApplicationInit):
            await extension.after_app_init(self)

    async def _call_on_shutdown_extensions(self) -> None:
        # Call module OnModuleDestroy hooks sequentially in reverse topological order (dependents first)
        for module in self._get_modules_for_triggering_extensions(reverse=True):
            for module_ext in self._extension_registry.get_module_extensions(module.target, OnModuleDestroy):
                await module_ext.on_module_destroy(module)

        for app_ext in self._extension_registry.get_application_extensions(OnApplicationShutdown):
            await app_ext.on_app_shutdown(self)

    def _get_modules_for_triggering_extensions(self, *, reverse: bool = False) -> Iterable[Module]:
        return reversed(self.registry.modules) if reverse else self.registry.modules



from __future__ import annotations

import abc
import enum
from abc import ABC
from typing import TYPE_CHECKING

from waku.eventsourcing._introspection import is_abstract

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar

    from waku.eventsourcing.contracts.event import StoredEvent
    from waku.eventsourcing.projection.checkpoint import Checkpoint
    from waku.messaging.contracts.event import IEvent

__all__ = [
    'ErrorPolicy',
    'ICatchUpProjection',
    'ICheckpointStore',
    'IProjection',
]


class IProjection(abc.ABC):
    projection_name: ClassVar[str]

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if is_abstract(cls):
            return
        if not getattr(cls, 'projection_name', None):
            msg = f'{cls.__name__} must define projection_name class attribute'
            raise TypeError(msg)

    @abc.abstractmethod
    async def project(self, events: Sequence[StoredEvent], /) -> None: ...


@enum.unique
class ErrorPolicy(enum.StrEnum):
    SKIP = enum.auto()
    STOP = enum.auto()


class ICatchUpProjection(IProjection, ABC):
    """Projection that processes events asynchronously via polling.

    At-least-once delivery: the checkpoint is saved *after* ``project()`` processes
    a batch, so a crash before checkpoint save causes re-delivery on restart.
    ``project()`` must be idempotent.

    Set ``event_types`` to filter which event types this projection receives.
    When ``None`` (default), all events are delivered.
    """

    event_types: ClassVar[Sequence[type[IEvent]] | None] = None

    async def on_skip(self, events: Sequence[StoredEvent], error: Exception) -> None:
        pass

    async def teardown(self) -> None:
        pass


class ICheckpointStore(abc.ABC):
    @abc.abstractmethod
    async def load(self, projection_name: str, /) -> Checkpoint | None: ...

    @abc.abstractmethod
    async def save(self, checkpoint: Checkpoint, /) -> None: ...



from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.contracts.pipeline import IPipelineBehavior, NextHandlerType
from waku.messaging.contracts.request import IRequest
from waku.messaging.events.handler import EventHandler, IEventHandler
from waku.messaging.impl import MessageBus
from waku.messaging.interfaces import IMessageBus, IPublisher, ISender
from waku.messaging.modules import MessagingConfig, MessagingExtension, MessagingModule
from waku.messaging.requests.handler import IRequestHandler, RequestHandler

__all__ = [
    'EventHandler',
    'EventT',
    'IEvent',
    'IEventHandler',
    'IMessageBus',
    'IPipelineBehavior',
    'IPublisher',
    'IRequest',
    'IRequestHandler',
    'ISender',
    'MessageBus',
    'MessagingConfig',
    'MessagingExtension',
    'MessagingModule',
    'NextHandlerType',
    'RequestHandler',
]



from __future__ import annotations

from waku.eventsourcing.contracts.stream import AnyVersion, Exact, NoStream, StreamExists, StreamId
from waku.eventsourcing.exceptions import ConcurrencyConflictError

__all__ = ['check_expected_version']


def check_expected_version(
    stream_id: StreamId,
    expected: Exact | NoStream | StreamExists | AnyVersion,
    current_version: int,
    *,
    exists: bool,
) -> None:
    match expected:
        case AnyVersion():
            return
        case NoStream():
            if exists:
                raise ConcurrencyConflictError(stream_id, -1, current_version)
        case StreamExists():
            if not exists:
                raise ConcurrencyConflictError(stream_id, 0, -1)
        case Exact(version=v):
            if v != current_version:
                raise ConcurrencyConflictError(stream_id, v, current_version)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from waku.di import Provider, many, scoped
from waku.messaging.events.map import EventMap
from waku.messaging.pipeline.map import PipelineBehaviorMap
from waku.messaging.requests.map import RequestMap

if TYPE_CHECKING:
    from collections.abc import Iterator

__all__ = ['MessageRegistry']


@dataclass(slots=True, kw_only=True)
class MessageRegistry:
    request_map: RequestMap = field(default_factory=RequestMap)
    event_map: EventMap = field(default_factory=EventMap)
    behavior_map: PipelineBehaviorMap = field(default_factory=PipelineBehaviorMap)

    def merge(self, other: MessageRegistry) -> None:
        self.request_map.merge(other.request_map)
        self.event_map.merge(other.event_map)
        self.behavior_map.merge(other.behavior_map)

    def freeze(self) -> None:
        self.request_map.freeze()
        self.event_map.freeze()
        self.behavior_map.freeze()

    def handler_providers(self) -> Iterator[Provider]:
        for entry in self.request_map.registry.values():
            yield scoped(entry.di_lookup_type, entry.handler_type)
        for entry in self.event_map.registry.values():
            yield many(entry.di_lookup_type, *entry.handler_types, collect=False)
        for entry in self.behavior_map.registry.values():
            yield many(entry.di_lookup_type, *entry.behavior_types, collect=False)

    def collector_providers(self) -> Iterator[Provider]:
        for entry in self.event_map.registry.values():
            yield many(entry.di_lookup_type, collect=True)
        for entry in self.behavior_map.registry.values():
            yield many(entry.di_lookup_type, collect=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from waku.eventsourcing.exceptions import UpcasterChainError

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from waku.eventsourcing.upcasting.interfaces import IEventUpcaster

__all__ = ['UpcasterChain']


class UpcasterChain:
    __slots__ = ('_chains',)

    def __init__(self, upcasters_by_type: Mapping[str, Sequence[IEventUpcaster]]) -> None:
        chains: dict[str, tuple[IEventUpcaster, ...]] = {}
        for event_type, upcasters in upcasters_by_type.items():
            sorted_upcasters = sorted(upcasters, key=lambda u: u.from_version)
            seen: set[int] = set()
            for u in sorted_upcasters:
                if u.from_version < 1:
                    msg = f'Invalid from_version {u.from_version} for event type {event_type!r}: must be >= 1'
                    raise UpcasterChainError(msg)
                if u.from_version in seen:
                    msg = f'Duplicate upcaster for event type {event_type!r} at from_version {u.from_version}'
                    raise UpcasterChainError(msg)
                seen.add(u.from_version)
            chains[event_type] = tuple(sorted_upcasters)
        self._chains = chains

    def upcast(self, event_type: str, data: dict[str, Any], schema_version: int) -> dict[str, Any]:
        upcasters = self._chains.get(event_type)
        if not upcasters:
            return data
        if schema_version > upcasters[-1].from_version:
            return data
        for u in upcasters:
            if u.from_version >= schema_version:
                data = u.upcast(data)
        return data



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import Generic, TypeAlias

from waku.messaging.contracts.request import RequestT, ResponseT

__all__ = [
    'IPipelineBehavior',
    'NextHandlerType',
]

NextHandlerType: TypeAlias = Callable[[RequestT], Awaitable[ResponseT]]


class IPipelineBehavior(ABC, Generic[RequestT, ResponseT]):
    """Interface for pipeline behaviors that wrap request handling."""

    @abstractmethod
    async def handle(self, request: RequestT, /, next_handler: NextHandlerType[RequestT, ResponseT]) -> ResponseT:
        """Handle the request and call the next handler in the pipeline.

        Args:
            request: The request to handle
            next_handler: Function to call the next handler in the pipeline

        Returns:
            The response from the pipeline
        """
        ...



from waku.eventsourcing.projection.sqlalchemy.store import SqlAlchemyCheckpointStore, make_sqlalchemy_checkpoint_store
from waku.eventsourcing.projection.sqlalchemy.tables import bind_checkpoint_tables

__all__ = [
    'SqlAlchemyCheckpointStore',
    'bind_checkpoint_tables',
    'make_sqlalchemy_checkpoint_store',
]



from waku.eventsourcing.store.in_memory import InMemoryEventStore
from waku.eventsourcing.store.interfaces import IEventReader, IEventStore, IEventWriter

__all__ = [
    'IEventReader',
    'IEventStore',
    'IEventWriter',
    'InMemoryEventStore',
]



from __future__ import annotations

from typing import TYPE_CHECKING

from waku.exceptions import WakuError

if TYPE_CHECKING:
    from waku.eventsourcing.contracts.stream import StreamId

__all__ = [
    'AggregateNotFoundError',
    'ConcurrencyConflictError',
    'ConflictingEventTypeError',
    'DuplicateAggregateNameError',
    'DuplicateEventTypeError',
    'DuplicateIdempotencyKeyError',
    'EventSourcingConfigError',
    'EventSourcingError',
    'PartialDuplicateAppendError',
    'ProjectionError',
    'ProjectionStoppedError',
    'RegistryFrozenError',
    'SnapshotConfigNotFoundError',
    'SnapshotMigrationChainError',
    'SnapshotTypeMismatchError',
    'StreamDeletedError',
    'StreamNotFoundError',
    'StreamTooLargeError',
    'UnknownEventTypeError',
    'UpcasterChainError',
]


class EventSourcingError(WakuError):
    pass


class EventSourcingConfigError(EventSourcingError):
    pass


class StreamNotFoundError(EventSourcingError):
    def __init__(self, stream_id: StreamId) -> None:
        self.stream_id = stream_id
        super().__init__(f'Stream {stream_id} not found')


class StreamDeletedError(EventSourcingError):
    def __init__(self, stream_id: StreamId) -> None:
        self.stream_id = stream_id
        super().__init__(f'Stream {stream_id} is deleted')


class AggregateNotFoundError(EventSourcingError):
    def __init__(self, aggregate_type: str, aggregate_id: str) -> None:
        self.aggregate_type = aggregate_type
        self.aggregate_id = aggregate_id
        super().__init__(f'{aggregate_type} with id {aggregate_id!r} not found')


class ConcurrencyConflictError(EventSourcingError):
    def __init__(self, stream_id: StreamId, expected_version: int, actual_version: int) -> None:
        self.stream_id = stream_id
        self.expected_version = expected_version
        self.actual_version = actual_version
        super().__init__(
            f'Concurrency conflict on stream {stream_id}: expected version {expected_version}, actual {actual_version}'
        )


class DuplicateAggregateNameError(EventSourcingError):
    def __init__(self, aggregate_name: str, repositories: list[type]) -> None:
        self.aggregate_name = aggregate_name
        self.repositories = repositories
        repo_names = ', '.join(r.__name__ for r in repositories)
        super().__init__(f'Duplicate aggregate name {aggregate_name!r} used by multiple repositories: {repo_names}')


class UnknownEventTypeError(EventSourcingError):
    def __init__(self, event_type_name: str) -> None:
        self.event_type_name = event_type_name
        super().__init__(f'Unknown event type: {event_type_name!r}')


class DuplicateEventTypeError(EventSourcingError):
    def __init__(self, event_type_name: str) -> None:
        self.event_type_name = event_type_name
        super().__init__(f'Event type {event_type_name!r} is already registered')


class ConflictingEventTypeError(EventSourcingError):
    def __init__(
        self,
        event_type: type,
        existing_name: str,
        existing_version: int,
        attempted_name: str,
        attempted_version: int,
    ) -> None:
        self.event_type = event_type
        self.existing_name = existing_name
        self.existing_version = existing_version
        self.attempted_name = attempted_name
        self.attempted_version = attempted_version
        if existing_name != attempted_name:
            detail = f'name {existing_name!r} → {attempted_name!r}'
        else:
            detail = f'version v{existing_version} → v{attempted_version}'
        super().__init__(f'Conflicting registration for event type {event_type.__name__!r}: {detail}')


class SnapshotTypeMismatchError(EventSourcingError):
    def __init__(self, stream_id: StreamId, expected_type: str, actual_type: str) -> None:
        self.stream_id = stream_id
        self.expected_type = expected_type
        self.actual_type = actual_type
        super().__init__(
            f'Snapshot type mismatch on stream {stream_id}: expected {expected_type!r}, got {actual_type!r}'
        )


class StreamTooLargeError(EventSourcingError):
    def __init__(self, stream_id: StreamId, max_length: int) -> None:
        self.stream_id = stream_id
        self.max_length = max_length
        super().__init__(
            f'Stream {stream_id} exceeds maximum length of {max_length} events. '
            f'Configure snapshots to reduce stream replay size.'
        )


class RegistryFrozenError(EventSourcingError):
    def __init__(self) -> None:
        super().__init__('Cannot register event types after registry is frozen')


class ProjectionError(EventSourcingError):
    pass


class ProjectionStoppedError(ProjectionError):
    def __init__(self, projection_name: str, cause: Exception) -> None:
        self.projection_name = projection_name
        self.cause = cause
        super().__init__(f'Projection {projection_name!r} stopped due to error: {cause}')


class DuplicateIdempotencyKeyError(EventSourcingError):
    def __init__(self, stream_id: StreamId, *, reason: str) -> None:
        self.stream_id = stream_id
        self.reason = reason
        super().__init__(f'Duplicate idempotency keys ({reason}) on stream {stream_id}')


class PartialDuplicateAppendError(EventSourcingError):
    def __init__(self, stream_id: StreamId, existing_count: int, total_count: int) -> None:
        self.stream_id = stream_id
        self.existing_count = existing_count
        self.total_count = total_count
        super().__init__(
            f'Partial duplicate append on stream {stream_id}: '
            f'{existing_count} of {total_count} idempotency keys already exist'
        )


class SnapshotConfigNotFoundError(EventSourcingError):
    def __init__(self, aggregate_name: str) -> None:
        self.aggregate_name = aggregate_name
        super().__init__(
            f'No snapshot config found for aggregate {aggregate_name!r}. '
            f'Provide snapshot=SnapshotOptions(...) via bind_aggregate() or bind_decider().'
        )


class SnapshotMigrationChainError(EventSourcingError):
    pass


class UpcasterChainError(EventSourcingError):
    pass



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, overload

if TYPE_CHECKING:
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.request import IRequest, ResponseT

__all__ = [
    'IMessageBus',
    'IPublisher',
    'ISender',
]


class ISender(abc.ABC):
    """Send requests through the messaging pipeline to be handled by a single handler."""

    @overload
    async def invoke(self, request: IRequest[None], /) -> None: ...

    @overload
    async def invoke(self, request: IRequest[ResponseT], /) -> ResponseT: ...

    @abc.abstractmethod
    async def invoke(self, request: IRequest[ResponseT], /) -> ResponseT:
        """In-process request/response. Always local, never routed externally."""

    @abc.abstractmethod
    async def send(self, request: IRequest[Any], /) -> None:
        """Fire-and-forget. Routable through outbox/transport. No response."""


class IPublisher(abc.ABC):
    """Publish events to be handled by multiple handlers."""

    @abc.abstractmethod
    async def publish(self, event: IEvent, /) -> None:
        """Fan-out to all subscribers. Routable through configured transport."""


class IMessageBus(ISender, IPublisher, abc.ABC):
    """Unified bus — inject this when you need both capabilities."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from waku.exceptions import WakuError

if TYPE_CHECKING:
    from waku.messaging.contracts.event import IEvent
    from waku.messaging.contracts.pipeline import IPipelineBehavior
    from waku.messaging.contracts.request import IRequest
    from waku.messaging.events.handler import IEventHandler
    from waku.messaging.requests.handler import IRequestHandler

__all__ = [
    'EventHandlerAlreadyRegistered',
    'ImproperlyConfiguredError',
    'MapFrozenError',
    'MessagingError',
    'PipelineBehaviorAlreadyRegistered',
    'RequestHandlerAlreadyRegistered',
    'RequestHandlerNotFound',
]


class MessagingError(WakuError):
    """Base exception for all messaging-related errors."""


class MapFrozenError(MessagingError):
    def __init__(self) -> None:
        super().__init__('Cannot modify map after it is frozen')


class ImproperlyConfiguredError(MessagingError):
    """Raised when messaging configuration is invalid."""


class RequestHandlerAlreadyRegistered(MessagingError, KeyError):  # noqa: N818
    """Raised when a request handler is already registered.

    Attributes:
        request_type: The type of request that caused the error.
        handler_type: The type of handler that was already registered.
    """

    def __init__(self, request_type: type[IRequest[Any]], handler_type: type[IRequestHandler[Any, Any]]) -> None:
        self.request_type = request_type
        self.handler_type = handler_type

    def __str__(self) -> str:
        return f'{self.request_type.__name__} already exists in registry with handler {self.handler_type.__name__}'


class RequestHandlerNotFound(MessagingError, TypeError):  # noqa: N818
    """Raised when a request handler is not found.

    Attributes:
        request_type: The type of request that caused the error.
    """

    def __init__(self, request_type: type[IRequest[Any]]) -> None:
        self.request_type = request_type

    def __str__(self) -> str:
        return f'Request handler for {self.request_type.__name__} request is not registered'


class EventHandlerAlreadyRegistered(MessagingError, KeyError):  # noqa: N818
    """Raised when an event handler is already registered.

    Attributes:
        event_type: The type of event that caused the error.
        handler_type: The type of handler that was already registered.
    """

    def __init__(self, event_type: type[IEvent], handler_type: type[IEventHandler[Any]]) -> None:
        self.event_type = event_type
        self.handler_type = handler_type

    def __str__(self) -> str:
        return f'{self.handler_type.__name__} already registered for {self.event_type.__name__} event'


class PipelineBehaviorAlreadyRegistered(MessagingError, KeyError):  # noqa: N818
    """Raised when a pipeline behavior is already registered.

    Attributes:
        request_type: The type of request that caused the error.
        behavior_type: The type of behavior that was already registered.
    """

    def __init__(self, request_type: type[IRequest[Any]], behavior_type: type[IPipelineBehavior[Any, Any]]) -> None:
        self.request_type = request_type
        self.behavior_type = behavior_type

    def __str__(self) -> str:
        return f'{self.behavior_type.__name__} already registered for {self.request_type.__name__} request'



from __future__ import annotations

import contextlib
import logging
from typing import TYPE_CHECKING

from sqlalchemy import text

from waku.eventsourcing.projection.lock.interfaces import IProjectionLock

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from sqlalchemy.ext.asyncio import AsyncEngine

__all__ = ['PostgresAdvisoryProjectionLock']

logger = logging.getLogger(__name__)

_LOCK_SQL = text('SELECT pg_try_advisory_lock(hashtext(:name))')
_UNLOCK_SQL = text('SELECT pg_advisory_unlock(hashtext(:name))')


class PostgresAdvisoryProjectionLock(IProjectionLock):
    """Session-level PostgreSQL advisory lock.

    Holds a database connection for the entire duration of the lock because
    ``pg_advisory_lock`` is bound to the session — releasing the connection
    releases the lock. For long-running projections consider
    :class:`PostgresLeaseProjectionLock` which only connects during heartbeats.

    Not compatible with PgBouncer in transaction-pooling mode.
    """

    def __init__(self, engine: AsyncEngine) -> None:
        self._engine = engine

    @contextlib.asynccontextmanager
    async def acquire(self, projection_name: str) -> AsyncGenerator[bool]:
        async with self._engine.connect() as conn:
            await conn.execution_options(isolation_level='AUTOCOMMIT')
            result = await conn.execute(_LOCK_SQL, {'name': projection_name})
            acquired = bool(result.scalar_one())

            if not acquired:
                yield False
                return

            logger.debug('Advisory lock acquired for %s', projection_name)
            try:
                yield True
            finally:
                try:
                    await conn.execute(_UNLOCK_SQL, {'name': projection_name})
                    logger.debug('Advisory lock released for %s', projection_name)
                except Exception:
                    logger.warning('Failed to release advisory lock for %s', projection_name, exc_info=True)



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence  # noqa: TC003  # Dishka needs runtime access
from datetime import UTC, datetime
from typing import TYPE_CHECKING, assert_never

import anyio

from waku.eventsourcing.contracts.event import EventEnvelope, IMetadataEnricher, StoredEvent
from waku.eventsourcing.contracts.stream import StreamPosition
from waku.eventsourcing.exceptions import (
    DuplicateIdempotencyKeyError,
    PartialDuplicateAppendError,
    StreamDeletedError,
    StreamNotFoundError,
)
from waku.eventsourcing.projection.interfaces import IProjection  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.serialization.registry import EventTypeRegistry  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.store._shared import enrich_metadata
from waku.eventsourcing.store._version_check import check_expected_version
from waku.eventsourcing.store.interfaces import IEventStore

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from waku.eventsourcing.contracts.stream import ExpectedVersion, StreamId

__all__ = ['InMemoryEventStore']


class InMemoryEventStore(IEventStore):
    def __init__(
        self,
        registry: EventTypeRegistry,
        projections: Sequence[IProjection] = (),
        enrichers: Sequence[IMetadataEnricher] = (),
    ) -> None:
        self._registry = registry
        self._streams: dict[str, list[StoredEvent]] = {}
        self._idempotency_keys: dict[str, set[str]] = {}
        self._deleted_streams: set[str] = set()
        self._global_position: int = 0
        self._lock = anyio.Lock()
        self._projections = projections
        self._enrichers = enrichers

    async def read_stream(
        self,
        stream_id: StreamId,
        /,
        *,
        start: int | StreamPosition = StreamPosition.START,
        count: int | None = None,
    ) -> list[StoredEvent]:
        async with self._lock:
            key = str(stream_id)
            if key not in self._streams:
                raise StreamNotFoundError(stream_id)
            events = self._streams[key]
            match start:
                case StreamPosition.START:
                    offset = 0
                case StreamPosition.END:
                    offset = max(len(events) - 1, 0)
                case int() as offset:
                    pass
                case _:  # pragma: no cover
                    assert_never(start)
            subset = events[offset:]
            if count is not None:
                subset = subset[:count]
            return list(subset)

    async def delete_stream(self, stream_id: StreamId, /) -> None:
        async with self._lock:
            key = str(stream_id)
            if key not in self._streams:
                raise StreamNotFoundError(stream_id)
            self._deleted_streams.add(key)

    async def read_all(
        self,
        *,
        after_position: int = -1,
        count: int | None = None,
        event_types: Sequence[str] | None = None,
    ) -> list[StoredEvent]:
        async with self._lock:
            all_events: list[StoredEvent] = []
            for key, stream_events in self._streams.items():
                if key not in self._deleted_streams:
                    all_events.extend(stream_events)
            all_events.sort(key=lambda e: e.global_position)

            type_set = frozenset(event_types) if event_types else None
            filtered = [
                e
                for e in all_events
                if e.global_position > after_position and (type_set is None or e.event_type in type_set)
            ]
            if count is not None:
                filtered = filtered[:count]
            return filtered

    async def stream_exists(self, stream_id: StreamId, /) -> bool:
        async with self._lock:
            key = str(stream_id)
            return key in self._streams and key not in self._deleted_streams

    async def global_head_position(self) -> int:
        async with self._lock:
            return self._global_position - 1

    async def read_positions(
        self,
        *,
        after_position: int,
        up_to_position: int,
    ) -> list[int]:
        async with self._lock:
            positions: list[int] = []
            for key, stream_events in self._streams.items():
                if key in self._deleted_streams:
                    continue
                positions.extend(
                    event.global_position
                    for event in stream_events
                    if after_position < event.global_position <= up_to_position
                )
            positions.sort()
            return positions

    async def append_to_stream(
        self,
        stream_id: StreamId,
        /,
        events: Sequence[EventEnvelope],
        *,
        expected_version: ExpectedVersion,
    ) -> int:
        async with self._lock:
            key = str(stream_id)
            if key in self._deleted_streams:
                raise StreamDeletedError(stream_id)
            stream = self._streams.get(key)
            current_version = len(stream) - 1 if stream is not None else -1

            if not events:
                check_expected_version(stream_id, expected_version, current_version, exists=stream is not None)
                return current_version

            dedup_version = self._check_idempotency(stream_id, events, current_version)
            if dedup_version is not None:
                return dedup_version

            check_expected_version(stream_id, expected_version, current_version, exists=stream is not None)

            if stream is None:
                stream = []
                self._streams[key] = stream
                is_new_stream = True
            else:
                is_new_stream = False

            stored_events: list[StoredEvent] = []
            base_global_position = self._global_position
            for envelope in events:
                position = len(stream)
                stored = StoredEvent(
                    event_id=uuid.uuid4(),
                    stream_id=stream_id,
                    event_type=self._registry.get_name(
                        type(envelope.domain_event)  # pyrefly: ignore[bad-argument-type]
                    ),
                    position=position,
                    global_position=self._global_position,
                    timestamp=datetime.now(UTC),
                    data=envelope.domain_event,
                    metadata=enrich_metadata(envelope.metadata, self._enrichers),
                    idempotency_key=envelope.idempotency_key,
                    schema_version=self._registry.get_version(
                        type(envelope.domain_event)  # pyrefly: ignore[bad-argument-type]
                    ),
                )
                stream.append(stored)
                stored_events.append(stored)
                self._global_position += 1

            stream_keys = self._idempotency_keys.setdefault(key, set())
            for envelope in events:
                stream_keys.add(envelope.idempotency_key)

            try:
                for projection in self._projections:
                    await projection.project(stored_events)
            except Exception:
                self._rollback_append(key, stream, events, base_global_position, is_new_stream=is_new_stream)
                raise

            return len(stream) - 1

    def _rollback_append(
        self,
        key: str,
        stream: list[StoredEvent],
        events: Sequence[EventEnvelope],
        base_global_position: int,
        *,
        is_new_stream: bool,
    ) -> None:
        del stream[len(stream) - len(events) :]
        if is_new_stream:
            del self._streams[key]
        self._global_position = base_global_position
        stream_keys = self._idempotency_keys.get(key)
        if stream_keys is not None:
            for envelope in events:
                stream_keys.discard(envelope.idempotency_key)
            if not stream_keys:
                del self._idempotency_keys[key]

    def _check_idempotency(
        self,
        stream_id: StreamId,
        events: Sequence[EventEnvelope],
        current_version: int,
    ) -> int | None:
        keys = [e.idempotency_key for e in events]
        unique_keys = set(keys)
        if len(unique_keys) != len(keys):
            raise DuplicateIdempotencyKeyError(stream_id, reason='duplicate keys within batch')

        existing = self._idempotency_keys.get(str(stream_id), set())
        found = unique_keys & existing

        if not found:
            return None

        if found == unique_keys:
            return current_version

        raise PartialDuplicateAppendError(stream_id, len(found), len(keys))



from __future__ import annotations

import warnings
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING, Final

from waku.extensions import AfterApplicationInit

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.application import WakuApplication
    from waku.validation import ValidationRule
    from waku.validation._errors import ValidationError

__all__ = [
    'ValidationContext',
    'ValidationExtension',
]


@dataclass(frozen=True, slots=True, kw_only=True)
class ValidationContext:
    app: WakuApplication


class ValidationExtension(AfterApplicationInit):
    __slots__ = ('rules', 'strict')

    def __init__(self, rules: Sequence[ValidationRule], *, strict: bool = True) -> None:
        self.rules = rules
        self.strict: Final = strict

    async def after_app_init(self, app: WakuApplication) -> None:
        context = ValidationContext(app=app)

        errors_chain = chain.from_iterable(rule.validate(context) for rule in self.rules)
        if errors := list(errors_chain):
            self._raise(errors)

    def _raise(self, errors: list[ValidationError]) -> None:
        if self.strict:
            msg = 'Validation error'
            raise ExceptionGroup(msg, errors)

        for error in errors:
            warnings.warn(str(error), stacklevel=3)



from __future__ import annotations

from collections.abc import AsyncGenerator, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import TYPE_CHECKING, TypeAlias, final

if TYPE_CHECKING:
    from waku.application import WakuApplication

__all__ = [
    'LifespanFunc',
    'LifespanWrapper',
]

LifespanFunc: TypeAlias = (
    Callable[['WakuApplication'], AbstractAsyncContextManager[None]] | AbstractAsyncContextManager[None]
)


@final
class LifespanWrapper:
    def __init__(self, lifespan_func: LifespanFunc) -> None:
        self._lifespan_func = lifespan_func

    @asynccontextmanager
    async def lifespan(self, app: WakuApplication) -> AsyncGenerator[None]:
        ctx_manager = (
            self._lifespan_func
            if isinstance(self._lifespan_func, AbstractAsyncContextManager)
            else self._lifespan_func(app)
        )
        async with ctx_manager:
            yield



from __future__ import annotations

from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import Generic, Self, TypeAlias

from typing_extensions import TypeVar

from waku.messaging.contracts.event import EventT, IEvent
from waku.messaging.events.handler import EventHandler
from waku.messaging.exceptions import EventHandlerAlreadyRegistered, MapFrozenError

__all__ = [
    'EventMap',
    'EventMapEntry',
    'EventMapRegistry',
]

_EventT = TypeVar('_EventT', bound=IEvent, default=IEvent)


@dataclass(frozen=True, slots=True)
class EventMapEntry(Generic[_EventT]):
    event_type: type[IEvent]
    di_lookup_type: type[EventHandler[_EventT]]
    handler_types: list[type[EventHandler[_EventT]]] = field(default_factory=list)

    @classmethod
    def for_event(cls, event_type: type[IEvent]) -> Self:
        di_lookup_type = EventHandler[event_type]  # type: ignore[valid-type]
        return cls(event_type=event_type, di_lookup_type=di_lookup_type)  # type: ignore[type-abstract]

    def add(self, handler_type: type[EventHandler[_EventT]]) -> None:
        if handler_type in self.handler_types:
            raise EventHandlerAlreadyRegistered(self.event_type, handler_type)
        self.handler_types.append(handler_type)


EventMapRegistry: TypeAlias = MutableMapping[type[IEvent], EventMapEntry[IEvent]]


class EventMap:
    def __init__(self) -> None:
        self._registry: EventMapRegistry = {}
        self._frozen = False

    def freeze(self) -> None:
        self._frozen = True

    @property
    def is_frozen(self) -> bool:
        return self._frozen

    def bind(self, event_type: type[EventT], handler_types: list[type[EventHandler[EventT]]]) -> Self:
        if self._frozen:
            raise MapFrozenError
        if event_type not in self._registry:
            self._registry[event_type] = EventMapEntry.for_event(event_type)

        entry = self._registry[event_type]
        for handler_type in handler_types:
            entry.add(handler_type)  # type: ignore[arg-type]
        return self

    def merge(self, other: EventMap) -> Self:
        if self._frozen:
            raise MapFrozenError
        for event_type, entry in other._registry.items():
            self.bind(event_type, entry.handler_types)
        return self

    @property
    def registry(self) -> EventMapRegistry:
        return self._registry

    def has_handlers(self, event_type: type[IEvent]) -> bool:
        return event_type in self._registry and len(self._registry[event_type].handler_types) > 0

    def get_handler_type(self, event_type: type[IEvent]) -> type[EventHandler[IEvent]]:
        return self._registry[event_type].di_lookup_type

    def __bool__(self) -> bool:
        return bool(self._registry)



from waku.messaging.events.handler import EventHandler, IEventHandler
from waku.messaging.events.publish import EventPublisher, GroupEventPublisher, SequentialEventPublisher

__all__ = [
    'EventHandler',
    'EventPublisher',
    'GroupEventPublisher',
    'IEventHandler',
    'SequentialEventPublisher',
]



from waku.eventsourcing.contracts.aggregate import EventSourcedAggregate, IDecider
from waku.eventsourcing.contracts.event import DataT, EventEnvelope, EventMetadata, IMetadataEnricher, StoredEvent
from waku.eventsourcing.contracts.stream import (
    AnyVersion,
    Exact,
    ExpectedVersion,
    NoStream,
    StreamExists,
    StreamId,
    StreamPosition,
)
from waku.eventsourcing.decider import (
    DeciderCommandHandler,
    DeciderRepository,
    DeciderVoidCommandHandler,
    SnapshotDeciderRepository,
)
from waku.eventsourcing.exceptions import (
    AggregateNotFoundError,
    ConcurrencyConflictError,
    ConflictingEventTypeError,
    DuplicateAggregateNameError,
    DuplicateEventTypeError,
    DuplicateIdempotencyKeyError,
    EventSourcingConfigError,
    EventSourcingError,
    PartialDuplicateAppendError,
    ProjectionError,
    ProjectionStoppedError,
    RegistryFrozenError,
    SnapshotConfigNotFoundError,
    SnapshotMigrationChainError,
    SnapshotTypeMismatchError,
    StreamDeletedError,
    StreamNotFoundError,
    StreamTooLargeError,
    UnknownEventTypeError,
    UpcasterChainError,
)
from waku.eventsourcing.handler import EventSourcedCommandHandler, EventSourcedVoidCommandHandler
from waku.eventsourcing.modules import (
    EventSourcingConfig,
    EventSourcingExtension,
    EventSourcingModule,
    EventType,
    EventTypeSpec,
    SnapshotOptions,
)
from waku.eventsourcing.projection.binding import CatchUpProjectionBinding
from waku.eventsourcing.projection.interfaces import ErrorPolicy, ICatchUpProjection, ICheckpointStore, IProjection
from waku.eventsourcing.projection.registry import CatchUpProjectionRegistry
from waku.eventsourcing.projection.runner import CatchUpProjectionRunner
from waku.eventsourcing.repository import EventSourcedRepository
from waku.eventsourcing.snapshot.repository import SnapshotEventSourcedRepository
from waku.eventsourcing.upcasting import (
    FnUpcaster,
    IEventUpcaster,
    UpcasterChain,
    add_field,
    noop,
    remove_field,
    rename_field,
    upcast,
)

__all__ = [
    'AggregateNotFoundError',
    'AnyVersion',
    'CatchUpProjectionBinding',
    'CatchUpProjectionRegistry',
    'CatchUpProjectionRunner',
    'ConcurrencyConflictError',
    'ConflictingEventTypeError',
    'DataT',
    'DeciderCommandHandler',
    'DeciderRepository',
    'DeciderVoidCommandHandler',
    'DuplicateAggregateNameError',
    'DuplicateEventTypeError',
    'DuplicateIdempotencyKeyError',
    'ErrorPolicy',
    'EventEnvelope',
    'EventMetadata',
    'EventSourcedAggregate',
    'EventSourcedCommandHandler',
    'EventSourcedRepository',
    'EventSourcedVoidCommandHandler',
    'EventSourcingConfig',
    'EventSourcingConfigError',
    'EventSourcingError',
    'EventSourcingExtension',
    'EventSourcingModule',
    'EventType',
    'EventTypeSpec',
    'Exact',
    'ExpectedVersion',
    'FnUpcaster',
    'ICatchUpProjection',
    'ICheckpointStore',
    'IDecider',
    'IEventUpcaster',
    'IMetadataEnricher',
    'IProjection',
    'NoStream',
    'PartialDuplicateAppendError',
    'ProjectionError',
    'ProjectionStoppedError',
    'RegistryFrozenError',
    'SnapshotConfigNotFoundError',
    'SnapshotDeciderRepository',
    'SnapshotEventSourcedRepository',
    'SnapshotMigrationChainError',
    'SnapshotOptions',
    'SnapshotTypeMismatchError',
    'StoredEvent',
    'StreamDeletedError',
    'StreamExists',
    'StreamId',
    'StreamNotFoundError',
    'StreamPosition',
    'StreamTooLargeError',
    'UnknownEventTypeError',
    'UpcasterChain',
    'UpcasterChainError',
    'add_field',
    'noop',
    'remove_field',
    'rename_field',
    'upcast',
]



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic

from typing_extensions import TypeVar

if TYPE_CHECKING:
    import uuid
    from datetime import datetime

    from waku.eventsourcing.contracts.stream import StreamId
    from waku.messaging.contracts.event import IEvent

DataT = TypeVar('DataT', bound='IEvent', default='IEvent')

__all__ = [
    'DataT',
    'EventEnvelope',
    'EventMetadata',
    'IMetadataEnricher',
    'StoredEvent',
]


@dataclass(frozen=True, slots=True, kw_only=True)
class EventMetadata:
    correlation_id: str | None = None
    causation_id: str | None = None
    extra: dict[str, Any] = field(default_factory=dict)


class IMetadataEnricher(abc.ABC):
    """Enriches event metadata before persistence."""

    @abc.abstractmethod
    def enrich(self, metadata: EventMetadata, /) -> EventMetadata: ...


@dataclass(frozen=True, slots=True, kw_only=True)
class EventEnvelope:
    domain_event: IEvent
    idempotency_key: str
    metadata: EventMetadata = field(default_factory=EventMetadata)

    def __post_init__(self) -> None:
        if not self.idempotency_key:
            msg = 'idempotency_key must not be empty'
            raise ValueError(msg)


@dataclass(frozen=True, slots=True, kw_only=True)
class StoredEvent(Generic[DataT]):
    event_id: uuid.UUID
    stream_id: StreamId
    event_type: str
    position: int
    global_position: int
    timestamp: datetime
    data: DataT
    metadata: EventMetadata
    idempotency_key: str
    schema_version: int = 1



from __future__ import annotations

from typing import TYPE_CHECKING

from waku.eventsourcing.projection.interfaces import ICheckpointStore

if TYPE_CHECKING:
    from waku.eventsourcing.projection.checkpoint import Checkpoint

__all__ = ['InMemoryCheckpointStore']


class InMemoryCheckpointStore(ICheckpointStore):
    def __init__(self) -> None:
        self._checkpoints: dict[str, Checkpoint] = {}

    async def load(self, projection_name: str, /) -> Checkpoint | None:
        return self._checkpoints.get(projection_name)

    async def save(self, checkpoint: Checkpoint, /) -> None:
        self._checkpoints[checkpoint.projection_name] = checkpoint



from __future__ import annotations

import abc
from typing import Any

__all__ = ['IEventUpcaster']


class IEventUpcaster(abc.ABC):
    from_version: int

    @abc.abstractmethod
    def upcast(self, data: dict[str, Any], /) -> dict[str, Any]: ...
